/*
    SPDX-FileCopyrightText: KDE Developers

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "kateabstractinputmodefactory.h"

KateAbstractInputModeFactory::KateAbstractInputModeFactory()
{
}

KateAbstractInputModeFactory::~KateAbstractInputModeFactory()
{
}
