v.setCursorPosition(0,18);
v.type(".");
v.type(".");
v.type(".");
v.type(".");
v.type(".");

v.setCursorPosition(1,32);
v.type(".");
v.type(".");

v.setCursorPosition(3,20);
v.type(".");
v.type(".");

v.setCursorPosition(4,44);
v.type(".");
v.type(".");

v.setCursorPosition(7,28);
v.type(".");
v.type(".");

v.setCursorPosition(9,12);
v.type(".");
v.type("ok");
v.type(".");
v.type(".");
v.type(".");
v.type(".");
