/*
    System-wide CMake indenter configuration.

    ATTENTION To override copy this file to ~/.kde4/share/apps/katepart/script/libraries
    and modify defined variables.

    This file is part of the Kate Project.
    SPDX-FileCopyrightText: Alex Turbov <i.zaufi@gmail.com>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

/**
 * Allow appending <tt>'{}'</tt> after \c '$' and trasformation of
 * <tt>'${}'</tt> to <tt>'$<>'</tt> and vise versa.
 */
var cmi_cfg_vgShortcut = true;
