v.setCursorPosition(11, document.lineLength(11));
v.enter();
v.type("file = read(f, String)");
v.enter();
v.type("print(file)");
v.enter();
v.type("end");
v.enter();
v.type("print(\"TEST2\")")
v.enter();
v.type("end");
v.enter();
v.type("print(a)");
v.enter();
v.type("end");
