v.setCursorPosition(0, document.lineLength(0));
v.enter();
v.type('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">');
v.enter();
v.enter();
v.type('<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">');
v.enter();
v.type('<head>');
v.enter();
v.type('<meta content="application/xhtml+xml; charset=utf-8" http-equiv="content-type" />');
v.enter();
v.type('<title>title</title>');
v.enter();
v.type('<link rel="stylesheet"');
v.enter();
v.type('type="text/css"');
v.enter();
v.type('media="screen"');
v.enter();
v.type('href="css/style.css" />');
v.enter();
v.type('<!--<link rel="alternate"');
v.enter();
v.type('type="application/rss+xml"');
v.enter();
v.type('title="Subscribe RSS-feed"');
v.enter();
v.type('href="feeds/feed.rss" />-->');
v.enter();
v.type('<style type="text/css">');
v.enter();
v.type('</style>');
v.enter();
v.type('<script type="text/javascript">');
v.enter();
v.type('<![CDATA[');
v.enter();
v.type(']]>');
v.enter();
v.type('</script>');
v.enter();
v.type('</head>');
v.enter();
v.enter();
v.type('<body>');
v.enter();
v.type('<p>');
v.enter();
v.type('XHTML is fun.');
v.enter();
v.type('</p>');
v.enter();
v.type('<p>And so is Open Source.</p>');
v.enter();
v.type('<ul>');
v.enter();
v.type('<li>test</li>');
v.enter();
v.type('<li>test2</li>');
v.enter();
v.type('</ul>');
v.enter();
v.type('</body>');
v.enter();
v.enter();
v.type('</html>');
