/*
    This file is part of the KDE libraries
    SPDX-FileCopyrightText: 2008 Niko Sams <niko.sams\gmail.com>

    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "codecompletiontestmodels.h"

#include "moc_codecompletiontestmodels.cpp"
