v.setCursorPosition(0,7);
v.type("|");
v.type("ok;");

v.setCursorPosition(1,8);
v.type("|");
v.type("ok;");

v.setCursorPosition(2,8);
v.type("|");
v.type("ok;");

v.setCursorPosition(3,9);
v.type("|");
v.type("ok;");

v.setCursorPosition(4,9);
v.type("|");
v.type("ok;");

v.setCursorPosition(5,10);
v.type("|");
v.type("ok;");

v.setCursorPosition(6,10);
v.type("|");
v.type("ok;");

v.setCursorPosition(7,9);
v.type("|");
v.type("ok;");

v.setCursorPosition(8,10);
v.type("|");
v.type("ok;");

