v.setCursorPosition(0,0);
v.align(new Range(new Cursor(0, 0), new Cursor(0, 0)));
v.setCursorPosition(1,0);
v.align(new Range(new Cursor(1, 0), new Cursor(1, 0)));
v.setCursorPosition(2,0);
v.align(new Range(new Cursor(2, 0), new Cursor(2, 0)));
v.setCursorPosition(3,0);
v.align(new Range(new Cursor(3, 0), new Cursor(3, 0)));
v.setCursorPosition(4,0);
v.align(new Range(new Cursor(4, 0), new Cursor(4, 0)));

v.setCursorPosition(7,0);
v.align(new Range(new Cursor(7, 0), new Cursor(7, 0)));
v.setCursorPosition(8,0);
v.align(new Range(new Cursor(8, 0), new Cursor(8, 0)));
v.setCursorPosition(9,0);
v.align(new Range(new Cursor(9, 0), new Cursor(9, 0)));
v.setCursorPosition(12,0);
v.align(new Range(new Cursor(12, 0), new Cursor(12, 0)));
