v.setCursorPosition(2,4);
v.type("private:");
v.enter();
v.type("void ok();");
v.enter();
v.type("public:");
v.enter();
v.type("void ok();");
v.enter();
v.type("protected:");
v.enter();
v.type("void ok();");
v.enter();
v.type("public slots:");
v.enter();
v.type("void ok();");
v.enter();
v.type("private Q_SLOTS:");
v.enter();
v.type("void ok();");
